include l4pre 

  Copyright GMD, L4.MEMCTR, 30 05 96, 20
 
 
;*********************************************************************
;******                                                         ******
;******         Memory Controller                               ******
;******                                                         ******
;******                                   Author:   J.Liedtke   ******
;******                                                         ******
;******                                   created:  05.03.91    ******
;******                                   modified: 30.05.96    ******
;******                                                         ******
;*********************************************************************
 
  public init_memctr
  public init_sigma0_1
  public grab_frame
  public phys_frames
 
  extrn create_kernel_including_task:near
  extrn map_ur_page_initially:near
  extrn physical_kernel_info_page:abs
  extrn dcod_start:byte
  extrn default_kdebug_end:byte
  extrn default_sigma0_end:byte
  extrn default_sigma1_end:byte
  extrn ktest_end:byte


 
include l4const
include uid
.listi
include adrspace
.xlisti
include tcb
include pagconst
include syscalls
include msg
include kpage


ok_for i486,pentium,ppro




                    align 4


phys_frames               dd    0

lowest_allocated_frame    dd    0


                    align 4


;----------------------------------------------------------------------------
;
;       grab frame / grab mem
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   size of grabbed area / -
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       phys_mem:EAX  begin of grabbed mem (4096 bytes / n*4096 bytes)
;
;----------------------------------------------------------------------------

  assume ds:codseg


grab_frame:

  push  ebx
  push  ecx
  push  edx
  push  esi
  push  edi
  push  ebp

  sub   edx,edx
  IFAE  esp,<offset tcb_space>
  CANDB esp,<offset tcb_space+size tcb_space>
        add   edx,PM
  FI
  mov   eax,[edx+lowest_allocated_frame]

  test  eax,eax
  jnz   initial_grab_frame


  sub   eax,eax
  mov   ecx,eax
  lea   edx,[eax-2]              ; w0 = FFFFFFFE

  log2  <%size physical_kernel_memory>

  mov   ebp,log2_*4+map_msg

  mov   esi,sigma0_task
  sub   edi,edi

  int   ipc

  test  al,ipc_error_mask
  CORNZ
  test  al,map_msg
  CORZ
  test  bl,fpage_grant
  CORZ
  shr   bl,2
  IFNZ  bl,log2_pagesize

        ke    <'-',0E5h,'0_err'>
  FI

  mov   eax,edx
  and   eax,-pagesize


  pop   ebp
  pop   edi
  pop   esi
  pop   edx
  pop   ecx
  pop   ebx
  ret



  icode


initial_grab_frame:

  sub   eax,pagesize

  IFB_  eax,MB1
        ke    '-memory_underflow'
  FI

  mov   [edx+lowest_allocated_frame],eax

  pop   ebp
  pop   edi
  pop   esi
  pop   edx
  pop   ecx
  pop   ebx
  ret



  icod  ends



;-----------------------------------------------------------------------
;
;       init memory controller
;
;-----------------------------------------------------------------------
; PRECONDITION:
;
;       protected mode, paging not yet enabled
;
;       disable interrupt
;
;       DS : R/W 0..4GB
;       CS : X/R 0..4GB, USE32
;
;-----------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX...EBP scratch
;
;----------------------------------------------------------------------------
 

 assume ds:codseg
 
  icode


 
init_memctr:

  mov   eax,offset dcod_start
  mov   ecx,ds:[kdebug_end+physical_kernel_info_page]
  IFZ   ecx,<offset default_kdebug_end>
  CANDA ecx,eax
        mov   eax,ecx
  FI
  mov   ecx,ds:[sigma0_ktask+physical_kernel_info_page].ktask_end
  IFZ   ecx,<offset default_sigma0_end>
  CANDA ecx,eax
        mov   eax,ecx
  FI
  mov   ecx,ds:[sigma1_ktask+physical_kernel_info_page].ktask_end
  IFZ   ecx,<offset default_sigma1_end>
  CANDA ecx,eax
        mov   eax,ecx
  FI
  mov   ecx,ds:[booter_ktask+physical_kernel_info_page].ktask_end
  IFZ   ecx,<offset ktest_end>
  CANDA ecx,eax
        mov   eax,ecx
  FI
  add   eax,pagesize-1
  and   eax,-pagesize

  mov   ds:[reserved_mem0+physical_kernel_info_page].mem_end,eax
  mov   ds:[reserved_mem0+physical_kernel_info_page].mem_begin,physical_kernel_info_page


  mov   eax,ds:[main_mem+physical_kernel_info_page].mem_end
  mov   ds:[reserved_mem1+physical_kernel_info_page].mem_begin,eax
  mov   ds:[reserved_mem1+physical_kernel_info_page].mem_end,eax
  mov   [lowest_allocated_frame],eax

  shr   eax,log2_pagesize
  mov   [phys_frames],eax

  ret

 




;-----------------------------------------------------------------------
;
;       init sigma 0 and 1
;
;
; PRECONDITION:
;
;       interrupts disabled
;
;-----------------------------------------------------------------------


 


xpush   macro reg

  sub   ecx,4
  mov   [ecx],reg

  endm






init_sigma0_1:

  mov   ecx,ds:[sigma0_ktask+physical_kernel_info_page+PM].ktask_stack

  mov   edi,physical_kernel_info_page
  xpush edi

  lea   ebx,ds:[sigma0_ktask+physical_kernel_info_page+PM]
  mov   [ebx].ktask_stack,ecx

  mov   eax,sigma0_task
  call  create_kernel_including_task

  call  init_sigma0_space

  mov   eax,sigma1_task
  lea   ebx,ds:[sigma1_ktask+physical_kernel_info_page+PM]
  call  create_kernel_including_task

  ret





;-----------------------------------------------------------------------
;
;       init sigma 0 address space ptabs
;
;-----------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   addr of first available page
;
;----------------------------------------------------------------------------
 

init_sigma0_space:

  sub   eax,eax
  mov   edi,ds:[reserved_mem0+physical_kernel_info_page].mem_begin
  call  map_ur_pages

  mov   eax,offset physical_kernel_info_page
  mov   esi,eax
  mov   dl,page_user_permit+page_present
  call  map_ur_page_initially

  sub   edi,edi
  xchg  [lowest_allocated_frame+PM],edi            ; turn off simple grabbing
  mov   ds:[reserved_mem1+physical_kernel_info_page+PM].mem_begin,edi

  mov   eax,ds:[reserved_mem0+physical_kernel_info_page].mem_end

  call  map_ur_pages

  ret


map_ur_pages:

  mov   esi,eax
  mov   ebx,pagesize
  DO
        mov   dl,page_user_permit+page_write_permit+page_present
        call  map_ur_page_initially
        add   eax,ebx
        add   esi,ebx
        cmp   esi,edi
        REPEATB
  OD

  ret


  icod  ends



  code ends
  end
