include l4pre 


  Copyright GMD, L4.PAGFAULT, 04 07 96, 160 K
 
 
;*********************************************************************
;******                                                         ******
;******         Page Fault Handler  (486)                       ******
;******                                                         ******
;******                                   Author:   J.Liedtke   ******
;******                                                         ******
;******                                   created:  24.02.90    ******
;******                                   modified: 04.07.96    ******
;******                                                         ******
;*********************************************************************
 

  public init_pagfault
  public page_fault_handler
 

  extrn invalid_access_error:near
  extrn map_system_shared_page:near
 
  extrn ipc_sc:near
  extrn tcb_fault:near
  extrn flush_tcb:near
  extrn push_ipc_state:near
  extrn pop_ipc_state:near
  extrn cancel_if_within_ipc:near
  extrn tunnel_to:near
  extrn define_idt_gate:near
  extrn exception:near
  extrn phys_frames:dword
  extrn hw_config_descriptor:byte
  extrn rom_frame_addr:abs

 
include l4const
include uid
.listi
include adrspace
.xlisti
include tcb
include cpucb
include intrifc
include schedcb
include syscalls
.listi
include pagconst
include pagmac
.xlisti
include msg



ok_for i486,pentium,ppro
 


  assume ds:codseg





;****************************************************************************
;*****                                                                  *****
;*****                                                                  *****
;*****        PAGFAULT    INITITIALIZATION                              *****
;*****                                                                  *****
;*****                                                                  *****
;****************************************************************************





;-----------------------------------------------------------------------
;
;       init page fault handling
;
;-----------------------------------------------------------------------
; PRECONDITION:
;
;       pm32
;
;       DS,ES linear space
;
;-----------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX...EBP scratch
;
;-----------------------------------------------------------------------
 
  icode


 
init_pagfault:

  mov   eax,offset page_fault_handler
  mov   bl,page_fault
  mov   bh,0
  call  define_idt_gate

  ret
 

  icod  ends



;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  kcode


;----------------------------------------------------------------------------
;
;                 Page Fault Handler
;
;----------------------------------------------------------------------------
;
;       Analyzes Page Faults and passes valid Page Faults to kernel.
;
;----------------------------------------------------------------------------
; Remark:
;
;       The linear addresses 1 MB ... 1 MB + 64 KB - 1 are aliased with
;       0 ... 64 KB - 1 to emulate 8086 wrap around.
;
;----------------------------------------------------------------------------



iret_   equ 0CFh




        align 16




XHEAD page_fault_pl0

 
  test  byte ptr ss:[esp+iret_eflags+4+2],(1 SHL (vm_flag-16)) 
  xret  nz,long 
 
  push  eax 

  test  esp,(size tcb-1) AND (-512)
  IFZ
  CANDA esp,<offset tcb_space>
  CANDB esp,<offset tcb_space+size tcb_space>
        ke    'esp < 512'
  FI

  mov   eax,ss:[esp+iret_eip+4+4] 

        mov   eax,cs:[eax]
                                          ; if PF happens at IRET (in PL0)
        IFZ   al,iret_                    ; new iret vector is dropped
                                          ; and faulting vector is taken
                                          ; instead. This ensures correct
                                          ; load of seg reg.
              mov   eax,ss:[esp+4]
              or    al,page_fault_from_user_level
              mov   ss:[esp+3*4+4+4],eax

              pop   eax
              add   esp,3*4+4             ; NZ !
              xret  ,long

              klign 16
        FI

        and   ah,NOT 7                    ;
        IFNZ  eax,0FF0040F6h              ; test byte ptr [reg],FF
        CANDNZ eax,0FF006080h             ; and  byte ptr [reg],FF
              pop   eax                   ; are skipped upon PF and result in C
              cmp   eax,eax               ;
              xret  ,long                 ; Z !
        FI 

        push  ebx
        push  ecx

        mov   ecx,cr3
        and   ecx,-pagesize
        mov   eax,cr2
        xpdir ebx,eax

        IFAE  eax,shared_table_base
        CANDB eax,shared_table_base+shared_table_size
              mov   ecx,ss:[kernel_proot]
        FI

        mov   ecx,ss:[(ebx*4)+ecx+PM]
        test  cl,page_present
        IFNZ
        and   ecx,-pagesize
        xptab eax,eax
        test  byte ptr ss:[(eax*4)+ecx+PM],page_present
        CANDNZ
              and   byte ptr ss:[esp+iret_eflags+4*4],NOT (1 SHL c_flag)
        ELSE_
              or    byte ptr ss:[esp+iret_eflags+4*4],1 SHL c_flag
        FI
        add   ss:[esp+iret_eip+4*4],4

        pop   ecx
        pop   ebx
        pop   eax
        add   esp,4

        iretd







  klign 16






page_fault_handler:

  cmp   ss:[esp+iret_cs+4],phys_mem_exec
  xc    z,page_fault_pl0,long


  ipre  ec_present

  mov   ebp,esp

  mov   ebx,-1
  IFNZ
        mov   ebx,[ebp+ip_eip]
  FI


  mov   edx,cr2

  mov   cl,byte ptr [ebp+ip_error_code]
  if    fpopn_write gt page_fault_due_to_write_bit
        shl   cl,fpopn_write - page_fault_due_to_write_bit
  endif
  if    fpopn_write lt page_fault_due_to_write_bit
        shr   cl,page_fault_due_to_write_bit - fpopn_write
  endif
  and   cl,page_fault_due_to_write

  and   dl,NOT 3
  or    dl,cl

  and   ebp,-size tcb

  IFB_  edx,<size virtual_space>,long

        mov   esi,[ebp+pager]
        mov   edi,[ebp+pager+4]

        IFNZ  [ebp+fine_state],locked_running

              sub   ecx,ecx
              sub   eax,eax
              mov   ebp,32*4+map_msg

              push  0
              push  phys_mem_exec
              push  offset ipcret
              jmp   ipc_sc
              ipcret:

              test  al,ipc_error_mask
              mov   al,page_fault
              jnz   exception

              ipost
        FI

        call  push_ipc_state

        IFNZ
              sub   eax,eax

              push  edx
              push  ebp
              mov   ebp,32*4+map_msg
              int   ipc
              pop   ebp
              pop   edx

              test  al,ipc_error_mask
              IFZ
              test__page_present edx
              CANDNC

                    call  pop_ipc_state

                    ipost
              FI
        FI
        jmp   cancel_if_within_ipc

  FI


  mov   al,page_fault
  cmp   ebx,-1
  jnz   exception

  mov   eax,edx

  cmp   eax,shared_table_base
  jb    short addressing_error

  cmp   eax,shared_table_base+shared_table_size-1
  jbe   shared_table_fault

  cmp   eax,offset iopbm
  jb    short addressing_error

  cmp   eax,offset iopbm+size iopbm-1
  jbe   own_iopbm_fault

  cmp   eax,offset com0_space
  jb    short addressing_error

  cmp   eax,offset com1_space+size com1_space-1
  jbe   com_space_write_fault






addressing_error:

internal_addressing_error:

  ke    '-inv_addr'













;----------------------------------------------------------------------------
;
;       shared table fault
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   faulting virtual address
;       EDX   = EAX
;       EBP   tcb write addr
;
;       DS    linear_space
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX,EBX,ECX,EDX,ESI,EDI        scratch
;
;----------------------------------------------------------------------------
;
;       PROC shared table fault (INT CONST address, attributes) :
;
;         IF kernel has ptab for this address CAND
;            actual task has no ptab for this address
;            THEN enter kernel ptab link into actual pdir
;            ELSE decode access {and enter into kernel pdir too}
;         FI
;
;       ENDPROC shared table fault
;
;----------------------------------------------------------------------------
; shared table INVARIANT:
;
;       all shared table ptabs are linked to kernel pdir
;
;----------------------------------------------------------------------------

  align 16


shared_table_fault:

  shr   eax,22
  lea   eax,[(eax*4)+PM]
  mov   edi,ds:[kernel_proot]
  mov   ebx,[eax+edi]
  test  bl,page_present
  IFNZ
        mov   edi,cr3
        and   edi,-pagesize
        xchg  [eax+edi],ebx
        test  bl,page_present
        IFZ
              ipost
        FI
  FI
  mov   eax,edx

  cmp   eax,offset tcb_space+size tcb_space
  jb    tcb_fault


  cmp   eax,offset pdir_space
  jb    addressing_error

  cmp   eax,offset pdir_space+size pdir_space
  jb    pdir_space_fault

  jmp   addressing_error




;----------------------------------------------------------------------------
;
;       own iopbm fault
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   faulting virtual address
;       EDX   = EAX
;       EBP   tcb write addr
;
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX,EBX,ECX,EDX,ESI,EDI        scratch
;




own_iopbm_fault:

  ke    'iopbm_fault'

  ret



;----------------------------------------------------------------------------
;
;       com space write fault
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   faulting virtual address
;       EDX   = EAX
;       EBP   tcb write addr
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX,EBX,ECX,EDX,ESI,EDI        scratch
;
;----------------------------------------------------------------------------
;
;       PROC com space write fault (addr) :
;
;         calc addr in dest task ;
;         ensure ptab existing and write mapped in dest task ;
;         copy entry .
;
;         ensure ptab existing and write mapped in dest task :
;           REP
;             IF NOT ptab existing
;               THEN map ptab ;
;                    enter new ptab into comspace
;             ELIF NOT write mapped in dest task
;               THEN user space read write fault (dest task, dest task addr, write)
;               ELSE LEAVE
;          PER .
;
;        ENDPROC com space write fault
;
;----------------------------------------------------------------------------

        align 4


com_space_write_fault:

  mark__ressource ebp,com_used

  mov   esi,[ebp+com_partner]               ; com partner is tcb address

  sub   eax,com0_base
  CORB
  IFAE  eax,MB8
        sub   eax,com1_base-com0_base
  FI
  sub   edx,com0_base
  mov   edi,edx
  shr   edi,23
  shr   edx,23-1
  mov   di,[(edi*2)+ebp+waddr0]
  shl   edi,20
  add   eax,edi


  DO
        lea___pdir ebx,esi
        xpdir ecx,eax
        mov   ebx,[(ecx*4)+ebx]
        and   bl,NOT page_user_permit
        mov   edi,ebx

        and   bl,page_present+page_write_permit
        IFZ   bl,page_present+page_write_permit
        and   ebx,-pagesize
        xptab ecx,eax
        mov   ebx,[(ecx*4)+ebx+PM]
        and   bl,page_present+page_write_permit
        CANDZ bl,page_present+page_write_permit

              mov   [(edx*4)+pdir+(offset com0_space SHR 20)],edi

              ipost
        FI

        push  esi

        mov   edi,ebp
        mov   ebp,[ebp+com_partner]
        call  tunnel_to

        add   byte ptr [eax],0

        xchg  edi,ebp
        call  tunnel_to

        mov   ebp,edi
        pop   esi
        REPEAT
  OD





  kcod ends
;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||





;----------------------------------------------------------------------------
;
;       pdir space fault
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   faulting virtual address within pdir_space
;       EDX   = EAX
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX,EBX,ECX,EDX,ESI,EDI        scratch
;
;----------------------------------------------------------------------------

        align 4


pdir_space_fault:

  sub   eax,offset pdir_space
  shr   eax,12
  load__proot eax,eax

  mov   esi,edx
  call  map_system_shared_page

  ipost












  code ends
  end
