include l4pre 


  Copyright GMD, L4.DISPATCH, 02 07 96, 81 K
 
;*********************************************************************
;******                                                         ******
;******         Dispatcher                                      ******
;******                                                         ******
;******                                   Author:   J.Liedtke   ******
;******                                                         ******
;******                                   created:  07.03.90    ******
;******                                   modified: 02.07.96    ******
;******                                                         ******
;*********************************************************************
 

 
  public init_dispatcher
  public init_schedcb
  public start_dispatch
  public dispatch
  public timer_int
  public dispatcher_open_tcb
  public dispatcher_close_tcb
  public induce_timeouts_at_present_waitees


  extrn switch_context:near
  extrn deallocate_ressources:near
  extrn reallocate_ressources:near
  extrn create_kernel_thread:near
  extrn tcb_write_fault:near
  extrn ipcman_wakeup_tcb:near
  extrn get_bottom_state:near
  extrn define_idt_gate:near
  extrn flush_tcb:near
  extrn physical_kernel_info_page:abs

 
include l4const
include uid
include adrspace
include tcb
include intrifc
include cpucb
.listi
include schedcb
.xlisti
include lbmac
include syscalls
include kpage
include pagconst


ok_for 486,pentium,ppro



        align 4


present_chain_version     dd 0





;----------------------------------------------------------------------------
;
;       init schedcb data area
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       DS,ES   linear space
;
;       interrupt & memory controller have to be already initialized
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX,EBX,ECX,EDX,EBP,ESI,EDI      scratch
;
;----------------------------------------------------------------------------

  assume ds:codseg

  icode



init_schedcb:

  mov   edi,offset system_clock_low
  mov   ecx,(size scheduler_control_block-(offset system_clock_low))/4
  sub   eax,eax
  cld
  rep   stosd

  mov   ds:[busy_actual],offset dispatcher_tcb

  ret


  icod  ends



;----------------------------------------------------------------------------
;
;       init dispatcher & dispatcher tcb
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       DS,ES   linear space
;
;       interrupt & memory controller have to be already initialized
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       EAX,EBX,ECX,EDX,EBP,ESI,EDI      scratch
;
;----------------------------------------------------------------------------

  assume ds:codseg


  icode



init_dispatcher:

  mov   ebp,esp
  and   ebp,-size tcb

  mov   [ebp+rem_timeslice],1         ; dispatcher ts will never reach 0 !

  mov   ds:[busy_actual],offset dispatcher_tcb

  mov   ds:[system_clock_low],1

  mov   bl,thread_switch
  mov   bh,3 SHL 5
  mov   eax,offset switch_sc
  call  define_idt_gate

  mov   bl,thread_schedule
  mov   bh,3 SHL 5
  mov   eax,offset thread_schedule_sc
  call  define_idt_gate

  ret




init_dispatcher_tcb:

  llinit present,ebx
  or    ds:[dispatcher_tcb+list_state],is_present

  linit busy
  mov   ds:[busy_actual],offset dispatcher_tcb

  linit interrupted
  linit soon_wakeup
  linit late_wakeup

  ret


  icod  ends



;----------------------------------------------------------------------------
;
;       dispatcher_open_tcb
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EBP   tcb write addr
;
;----------------------------------------------------------------------------


dispatcher_open_tcb:
 
  pushad
  pushfd

  cli

  IFZ   ebp,<offset dispatcher_tcb>
        call  init_dispatcher_tcb
  FI


  inc   [present_chain_version+PM]

  test  [ebp+list_state],is_present
  IFZ
        lea   edx,[ebp+present_llink]       ; Attention: may already linked into
        mov   ecx,offset present_root
        llins edx,ecx,eax,present           ;            the present chain by
  FI                                        ;            concurrent tcb faults



  test  [ebp+fine_state],nwake
  IFNZ
        mov   cl,ds:[system_clock_high]
        mov   [ebp+wakeup_high],cl
        test  [ebp+fine_state],nbusy
        IFZ
        test  [ebp+list_state],is_busy
        CANDZ
              lins ebp,edx,busy
        FI
  FI

  popfd
  popad
  ret


;----------------------------------------------------------------------------
;
;       dispatcher close tcb
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EBP   tcb write addr
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       tcb eliminated from all dispatcher lists
;
;----------------------------------------------------------------------------


dispatcher_close_tcb:
 
  pushad
  pushfd

  cli

  inc   [present_chain_version+PM]          ; aborts concurrent parsing of
                                            ; the present chain
  mov   ebx,esp
  and   ebx,-size tcb

  test  [ebp+list_state],is_present
  IFNZ
        lea   ecx,[ebp+present_llink]
        lldel ecx,edx,eax,present
  FI

  mov   edx,offset busy_link
  mov   cl,is_busy
  call  delete_from_single_linked_list

  IFZ   ds:[busy_actual],ebp
        mov   ds:[busy_actual],offset dispatcher_tcb
  FI

  mov   edx,offset interrupted_link
  mov   cl,is_interrupted
  call  delete_from_single_linked_list

  mov   edx,offset late_wakeup_link
  mov   cl,is_late_wakeup
  call  delete_from_single_linked_list

  mov   edx,offset soon_wakeup_link
  mov   cl,is_soon_wakeup
  call  delete_from_single_linked_list

  btr   [ebp+wakeup_low],31
  IFC
        mov   al,ds:[system_clock_high]
        mov   [ebp+wakeup_high],al
  FI

;;lno___thread edx,ebp
;;test  [ebp+fine_state],nbusy
;;IFZ
;;      call  signal_scheduler_reactivation
;;ELSE_
;;      mov   eax,ebp
;;      test  [eax+fine_state],nwake
;;      IFZ
;;            mov   ecx,[eax+wakeup_low]
;;            mov   dx,[eax+wakeup_high]
;;            call  signal_scheduler_wakeup
;;      FI
;;      test  [eax+fine_state],nwait+nclos
;;      IFZ
;;            mov   ecx,[eax+waiting_for]
;;            lno___thread ecx,ecx
;;            call  signal_scheduler_waitfor
;;      FI
;;FI

  popfd
  popad
  ret






;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
  kcode




;----------------------------------------------------------------------------
;
;       delete from single linked list
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EBP   tcb to be deleted, write addr
;       EDX   list offset
;       CL    list mask
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       tcb deleted from list
;
;----------------------------------------------------------------------------



delete_from_single_linked_list:

  test  [ebp+list_state],cl
  IFNZ
        push  esi
        push  edi

        mov   edi,offset dispatcher_tcb
        klign 16
        DO
              mov   esi,edi
              mov   edi,[edi+edx]
              test  edi,edi
              EXITZ
              cmp   edi,ebp
              REPEATNZ

              not   cl
              and   [edi+list_state],cl
              mov   edi,[edi+edx]
              mov   [esi+edx],edi
        OD

        pop   edi
        pop   esi
  FI

  ret



;----------------------------------------------------------------------------
;
;       dispatcher thread
;
;----------------------------------------------------------------------------
;
;       REP
;         enable interrupts ;
;         disable interrupts ;
;         IF interrupted threads stack is empty
;           THEN get thread from busy que
;           ELSE pop thread from interrupted threads stack
;         FI ;
;         IF thread found
;           THEN switch to thread
;         FI
;       PER .
;
;----------------------------------------------------------------------------
; Remark: The dispatcher runs on its own thread, but for sake of efficiency
;         no complete switch_context (only a temporary stack switch) is used
;         for switching from any other thread to the dispatcher. Whenever a
;         new thread is picked up by the dispatcher or by a hardware interrupt
;         the last threads stackpointer is restored an a complete context
;         switch from the last to the new thread is executed. (Note that the
;         dispatcher thread can run in any address space.)
;         Pros:
;             1. Only one instead of two switches are necessary for dispatch,
;                especially at most one switching of address space.
;             2. If there is only one thread (and the dispatcher) busy in the
;                moment, address space switch and deallocate/allocate ressources
;                (numeric coprocesor) can be omitted.
;
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EBP         tcb write address of actual thread ( <> dispatcher ! )
;
;       DS, ES      linear space
;
;       interrupts disabled
;
;----------------------------------------------------------------------------





start_dispatch:

  mov   ebx,dispatcher_tcb
  jmp   short dispatch




  align 16



dispatch:
 
  mov   [ebx+thread_esp],esp
  mov   esp,offset dispatcher_tcb+size pl0_stack-size int_pm_stack

  sti                                 ; ensuring 'cli ; REP dispatch PER'
  nop                                 ; to permit interrupts
  nop                                 ; due to 486 bug (1 cycle enabled too short)
  cli

  movl__root eax,interrupted
  klign 16
  DO
        movl__next eax,edx,interrupted
        EXITZ
 
        mov   ebp,eax
        ldel  eax,edx,interrupted
        test  [ebp+fine_state],nbusy
        REPEATNZ

        IFNZ  ebp,ebx
        CANDG [ebp+rem_timeslice],0
              mov   esp,[ebx+thread_esp]
              jmp   switch_context
        FI

        mark__busy ebp
        REPEAT
  OD


  mov   ebp,ds:[busy_actual]
  DO
        movl__next ebp,edx,busy
        IFNZ
              test  [ebp+fine_state],nbusy
              IFNZ
                    ldel  ebp,edx,busy
                    REPEAT

                    klign 16
              FI

              mov   al,[ebp+timeslice]
              mov   [ebp+rem_timeslice],al

              mov   ds:[busy_actual],ebp

              mov   esp,[ebx+thread_esp]
              cmp   ebp,ebx
              jnz   switch_context
              ret

        FI

        movl__root ebp,busy
        mov   ds:[busy_actual],ebp

        sti
        nop
        nop
        cli

        REPEAT
  OD






;----------------------------------------------------------------------------
;
;       switch_sc
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       ESI   0 : dispatch
;
;       ESI   <>0 : donate, thread id (low)
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       REG                     scratch
;
;----------------------------------------------------------------------------



  align 16


switch_sc:
 
  tpre  switch_code,ds

  mov   ebp,esp
  and   ebp,-size tcb
  mov   ebx,ebp

  mark__busy ebp

  mov   al,[ebp+coarse_state]
  and   al,nblocked+ndead
  cmp   al,nblocked+ndead
  xc    nz,sw_block

  push  offset switch_ret

  lea___tcb ebp,esi
  cmp   ebp,dispatcher_tcb
  jbe   dispatch

  IFNZ  ebp,ebx
  test__page_writable ebp
  CANDNC

        mov   al,[ebp+fine_state]
        test  al,nbusy
        jz    switch_context

        and   al,NOT nwake
        IFZ   al,closed_wait
              cmp   [ebp+waiting_for],0
              jz    switch_context
        FI
  FI
  jmp   dispatch



  align 4



XHEAD sw_block

  bt    [esp+ip_eflags+4],vm_flag
  CORC
  test  byte ptr [esp+ip_cs+4],11b
  IFNZ
        or    [ebp+fine_state],nbusy
  FI
  xret






  align 16


switch_ret:

  tpost eax,ds,es







;----------------------------------------------------------------------------
;
;       timer interrupt
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       <ESP>    INTR return vector
;
;       interrupts disabled
;
;----------------------------------------------------------------------------
;
;       PROC timer interrupt :
;
;         pulse counter DECR thousand div milliseconds per pulse ;
;         IF pulse counter <= 0
;           THEN pulse counter INCR pulses per second ;
;                timer tick ;
;                IF end of timeslice
;                  THEN mark busy (myself)
;                FI ;
;                inspect wakeup lists ;
;                IF wakeup pending COR end of timeslice
;                  THEN IF myself in kernel mode
;                         THEN mark pending dispatch
;                         ELSE dispatch
;                       FI
;                FI
;         FI .
;
;         delta t : milliseconds per pulse .
;
;         timer tick :
;           increment system clocks ;
;           cpu clock (me) INCR delta t ;
;           remaining timeslice (me) := max (remaining timeslice - delta t, 0) .
;
;         increment system clocks :
;           system clock offset INCR delta t ;
;           propagate new clock to ipcman ;
;           IF ready threads = 0
;             THEN idle clock INCR delta t
;             ELSE ready clock INCR (ready threads * delta t) ;
;                  IF kernel active
;                    THEN kernel clock INCR delta t
;                  FI
;           FI .
;
;         inspect wakeup lists :
;                  IF system clock MOD 1024 = 0
;                    THEN parse all present tcbs
;                  FI ;
;                  IF system clock MOD 128 = 0
;                    THEN parse mid term wakeup list
;                  FI ;
;                  IF system clock MOD 4 = 0
;                    THEN parse short term wakeup list
;                  FI .
;
;
;         parse short term wakeup list :
;           actual := first (short term wakeup) ;
;           WHILE actual <> nil REP
;             abort wakeup handling if necessary ;
;             IF fine state (actual).wait
;               THEN IF wakeup (actual) <= system clock offset
;                      THEN push interrupted (myself) ;  {happens only once!}
;                           remaining timeslice (actual) := intr timeslice length ;
;                           push interrupted (actual)
;                           delete from short term wakeup list (actual)
;                    FI
;               ELSE delete from short term wakeup list (actual)
;             FI
;             actual := next (short term wakeup)
;           PER .
;
;         parse mid term wakeup list :
;           actual := first (mid term wakeup) ;
;           WHILE actual <> nil REP
;             abort wakeup handling if necessary ;
;             IF fine state (actual).wait
;               THEN IF wakeup (actual) <= system clock offset + 128
;                      THEN delete from mid term wakeup list (actual)
;                           insert into short term wakeup list (actual)
;                    FI
;               ELSE delete from mid term wakeup list (actual)
;             FI
;             actual := next (mid term wakeup)
;           PER .
;
;         parse long term wakeup list :
;           actual := first (present) ;
;           WHILE actual <> nil REP
;             abort wakeup handling if necessary ;
;             IF fine state (actual).wait
;               THEN IF wakeup (actual) <= system clock offset + 128
;                      THEN insert into mid term wakeup list (actual)
;                    FI
;             FI
;             actual := next (present)
;           PER .
;
;----------------------------------------------------------------------------

  align 16



timer_int:

  ipre  fault
 
  reset_timer_intr

  mov   ebx,esp
  and   ebx,-size tcb

  mov   esi,offset system_clock_low
  mov   edi,offset user_clock+physical_kernel_info_page+PM

  sub   [esi+pulse_counter-offset system_clock_low],thousand_div_millis
  IFBE  ,,long

        add   [esi+pulse_counter-offset system_clock_low],pulses_per_second
        mov   edx,millis_per_pulse
        mov   ecx,micros_per_pulse

        add   dword ptr ds:[esi],edx
        xc    c,inc_system_clock_high

        add   dword ptr ds:[edi],ecx
        xc    c,inc_user_clock_high

        add   [ebx+cpu_clock_low],edx
        xc    c,inc_cpu_clock_high

        IFNZ  ebx,dispatcher_tcb
              sub   [ebx+rem_timeslice],dl
        FI


        sub   edi,edi

        sub   ds:[late_wakeup_count],dl
        xc    c,late_wakeup,long

        sub   ds:[soon_wakeup_count],dl
        xc    c,soon_wakeup


        test  edi,edi
        CORNZ
        IFLE  [ebx+rem_timeslice],0

              cmp   ebx,dispatcher_tcb
              jz    dispatch

              mark__busy ebx

              mov   al,[ebx+coarse_state]
              and   al,nblocked+ndead
              cmp   al,nblocked+ndead
              IFNZ
                    or   [ebx+fine_state],nbusy
              FI
              call  dispatch
        FI
  FI

  ipost


XHEAD inc_system_clock_high

  inc   [esi+system_clock_high-offset system_clock_low]
  xret


XHEAD inc_user_clock_high

  inc   dword ptr ds:[edi+4]
  xret


XHEAD inc_cpu_clock_high

  inc   [ebx+cpu_clock_high]
  xret





XHEAD soon_wakeup

  mov   ds:[soon_wakeup_count],soon_wakeup_interval

  movl__root ebp,soon_wakeup
  mov   eax,ds:[system_clock_low]
  DO
        movl__next ebp,ecx,soon_wakeup
        xret  z

        test  [ebp+fine_state],nwake
        IFZ
              cmp   [ebp+wakeup_low],eax
              REPEATG

              mov   edi,ebp
              IFNZ  ebx,dispatcher_tcb
                    push_interrupted_thread ebx
              FI
              call  ipcman_wakeup_tcb
              push_interrupted_thread ebp
        FI
        ldel  ebp,ecx,soon_wakeup
        REPEAT
  OD





  kcod ends

;||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||



XHEAD   late_wakeup

  mov   ds:[late_wakeup_count],late_wakeup_interval

  sub   ds:[late_late_wakeup_count],1
  xc    c,late_late_wakeup,long

  mov   eax,ds:[system_clock_low]
  add___eax late_wakeup_interval
  movl__root ebp,late_wakeup
  clign 16
  DO
        movl__next ebp,ecx,late_wakeup
        xret  z,long
 
        test  [ebp+fine_state],nwake
        IFZ
              cmp   [ebp+wakeup_low],eax
              REPEATG

              test  [ebp+list_state],is_soon_wakeup
              IFZ
                    lins  ebp,esi,soon_wakeup
              FI
        FI
        ldel  ebp,ecx,late_wakeup
        REPEAT
  OD






XHEAD   late_late_wakeup

  mov   ds:[late_late_wakeup_count],late_late_wakeup_interval/late_wakeup_interval
  mov   eax,ds:[system_clock_low]
  add   eax,late_late_wakeup_interval
  push  edx
  sub   dl,dl
  mov   esi,ds:[present_chain_version+PM]

  movll_root ecx,present
  clign 16
  DO

        sub   dl,4
        xc    c,permit_interrupts

        movll_next ecx,present
        EXITZ
 
        lea   ebp,[ecx-present_llink]

        test  [ecx+fine_state-offset present_llink],nwake
        REPEATNZ

        test  [ebp+list_state],is_late_wakeup
        REPEATNZ

        cmp   [ebp+wakeup_low],eax
        REPEATG

        lins  ebp,esi,late_wakeup
        REPEAT

late_late_wakeup_od:

  OD
  pop   edx

  xret  ,long




XHEAD permit_interrupts

  sti
  nop
  nop
  cli
  cmp   esi,ds:[present_chain_version+PM]
  xret  z

  mov   ds:[late_late_wakeup_count],10
  jmp   late_late_wakeup_od




 
;----------------------------------------------------------------------------
;
;       thread schedule sc
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   param word
;       EBX   ext preempter (low)  / 0
;       EBP   ext preempter (high) / 0
;       ESI   thread id (low)
;       EDI   thread id (high)
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;
;
;----------------------------------------------------------------------------

schedule_failed:

  sub   eax,eax
  dec   eax

  tpost ebx




thread_schedule_sc:

  tpre  trap2,ds,es

  mov   edx,esi
  lea___tcb esi

  cmp   [esi+myself],edx
  IFZ
        cmp   [esi+myself+4],edi
  FI
  jnz   schedule_failed


        mov   edi,esp
        and   edi,-size tcb

        push  eax
        push  ebx

        mov   dl,[esi+timeslice]
        shl   edx,4
        add   edx,10
        shl   edx,8

        call  get_bottom_state
        mov   dl,0F0h
        IFZ   al,running
              mov   dl,00h
        FI
        IFZ   al,locked_running
              mov   dl,40h
        FI
        IFZ   al,locked_waiting
              mov   dl,80h
        FI
        test  al,nwait
        IFZ   al,closed_wait
              mov   dl,0C0h
        FI
        IFZ   al,polling
              mov   dl,0D0h
        FI
        IFB_  dl,0C0h
        CANDNZ [esi+state_sp],0
              mov   ebx,[esi+waiting_for]
              IFZ   ebx,[esi+pager]
                    add   dl,10h
              ELIFZ ebx,[esi+int_preempter]
                    add   dl,20h
              ELIFZ ebx,[esi+ext_preempter]
                    add   dl,30h
              FI
        FI

        pop   ebx
        pop   eax

        shl   edx,12
        mov   dl,[esi+prio]

        mov   ch,[edi+max_controlled_prio]

  cmp   dl,ch
  ja    schedule_failed


        IFNZ  eax,-1
        CANDBE al,ch

              mov   [esi+prio],al
              mov   ecx,eax
              shr   ecx,20
              and   cl,0Fh
              add   cl,2
              shr   eax,cl
              shr   eax,cl
              IFA   eax,127
                    mov   al,127
              FI
              mov   [esi+timeslice],al
        FI
        mov   eax,edx

        IFNZ  ebx,-1
              xchg  ebx,[esi+ext_preempter]
              xchg  ebp,[esi+ext_preempter+4]
        ELSE_
              mov   ebx,[esi+ext_preempter]
              mov   ebp,[esi+ext_preempter+4]
        FI

        push  eax

        mov   edi,1000

        sub   ecx,ecx
        test  [esi+fine_state],nwake
        IFZ
              dec   ecx
              mov   eax,[esi+wakeup_low]
              sub   eax,ds:[system_clock_low]
              mul   edi
              mov   al,dl
              test  eax,0FF0000FFh
              IFZ
                    rol   eax,16
                    add   cl,80h
              FI
              test  al,al
              IFZ
                    rol   eax,8
                    add   cl,40h
              FI
              test  al,0F0h
              IFZ
                    shl   eax,4
                    add   cl,20h
              FI
              test  al,0C0h
              IFZ
                    shl   eax,2
                    add   cl,10h
              FI
              mov   ch,cl
              shl   ecx,16
        FI

        mov   eax,[esi+cpu_clock_low]
        mul   edi
        add   edx,ecx
        movzx ecx,[esi+cpu_clock_high]
        imul  ecx,edi
        add   edx,ecx
        mov   ecx,eax

        pop   eax


        mov   edi,[esi+waiting_for+4]
        mov   esi,[esi+waiting_for]


  tpost ,ds,es










;----------------------------------------------------------------------------
;
;       induce timeouts at present waitees
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       ESI   thread id (low)
;       EDI   thread id (high)
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       all present threads waiting for <ESI,EDI> got receive timeout
;
;----------------------------------------------------------------------------

        align 16



XHEAD permit_interrupts_while_inducing_timeouts

  mov   eax,[present_chain_version+PM]
  sti
  mov   cl,16
  mov   cl,cl                   ; due to 486 bug (1 cycle enabled too short)
  cli
  cmp   [present_chain_version+PM],eax
  xret  z

  pop   eax                     ; restart induce_timeouts_at_present_waitees
  popfd                         ; if someone changed the present chain by
  popad                         ; interrupt



        clign 4




induce_timeouts_at_present_waitees:
 
  pushad
  pushfd

  cli
  mov   cl,16

  movll_root ebx,present
 
  clign 16
  DO
        movll_next ebx,present
        EXITZ

        dec   cl
        xc    z,permit_interrupts_while_inducing_timeouts

        lea   edx,[ebx-offset present_llink-offset succ]
 
        cmp   [edx+waiting_for],esi
        REPEATNZ
        cmp   [edx+waiting_for+4],edi
        REPEATNZ

        test  [edx+coarse_state],ndead
        REPEATZ

        mov   al,[edx+fine_state]
        test  al,npoll
        IFNZ
              and   al,nwait+nlock+nclos
              cmp   al,nlock
              REPEATNZ
        FI

        mov   ebp,edx
        call  ipcman_wakeup_tcb
        REPEAT
  OD


  popfd
  popad
  ret










        code ends
        end

