include l4pre 


  Copyright GMD, L4.INTCTR, 12 07 96, 56
 
 
;*********************************************************************
;******                                                         ******
;******         Interrupt Controller                            ******
;******                                                         ******
;******                                   Author:   J.Liedtke   ******
;******                                                         ******
;******                                   created:  19.07.87    ******
;******                                   modified: 12.07.96    ******
;******                                                         ******
;*********************************************************************


 
  public init_intctr
  public define_idt_gate
  public emu_lidt_eax
  public exception
 
 
  extrn intr_from_v86:near
  extrn shutdown_thread:near
  extrn machine_check_exception:near
  extrn physical_kernel_info_page:abs
 
 
include l4const
include adrspace
include uid
include tcb
include pagmac
.listi
include intrifc
include syscalls
.xlisti
include kpage
 

ok_for 486,pentium,ppro
 
 
  icode

 
idtvec  dw size idt-1
        dd offset idt
 
 
        align 4
 
 
 
r32               equ  0100000000010000b
rw32              equ  0100000000010010b
r16               equ             10000b
rw16              equ             10010b
 
x32               equ  0100000000011000b
xr32              equ  0100000000011010b
x16               equ             11000b
xr16              equ             11010b
 
ldtseg            equ  2
taskgate          equ  5
tsseg             equ  9
callgate          equ  0Ch
intrgate          equ  0Eh
trapgate          equ  0Fh
 

ar_byte        record dpresent:1,dpl:2,dtype:4,daccessed:1

d_bit          equ 22

 
;---------------------------------------------------------------------------
;
;        descriptor privilege levels codes
;
;---------------------------------------------------------------------------
 
dpl0    equ 0 shl 5
dpl1    equ 1 shl 5
dpl2    equ 2 shl 5
dpl3    equ 3 shl 5



;-----------------------------------------------------------------------
;
;       init interrupt controller
;
;-----------------------------------------------------------------------
; PRECONDITION:
;
;       protected mode
;
;       paging enabled, adrspace established
;
;       disable interrupt
;
;       DS,ES  linear space
;
;-----------------------------------------------------------------------
; POSTCONDITION:
;
;       IDT     initialized
;       IDTR    initialized
;
;       EAX...EBP scratch
;
;----------------------------------------------------------------------
 

 assume ds:codseg
 
 
init_intctr:

  sub   eax,eax
  mov   edi,offset idt
  mov   ecx,size idt/4
  cld
  rep   stosd

  mov   bl,0
  mov   esi,offset initial_idt+PM
  DO
        mov   eax,[esi]
        mov   bh,[esi+4]
        call  define_idt_gate
        inc   bl
        add   esi,5
        cmp   esi,offset end_of_initial_idt+PM
        REPEATB
  OD

  lidt  pword ptr ds:[idtvec+PM]

  ret


 
 
        align 4


initial_idt    dd offset divide_error_handler
               db dpl3
               dd offset initial_debug_exception_handler
               db dpl0
               dd 0
               db dpl0
               dd offset breakpoint_handler
               db dpl3
               dd offset overflow_handler
               db dpl3
               dd offset bound_check_handler
               db dpl3
               dd offset invalid_opcode_handler
               db dpl0
               dd 0
               db dpl0
               dd offset double_fault_handler
               db dpl0
               dd 0
               db dpl0
               dd 0
               db dpl0
               dd offset seg_not_present_handler
               db dpl0
               dd offset stack_exception_handler
               db dpl0
               dd 0
               db dpl0
               dd 0
               db dpl0
               dd 0
               db dpl0
               dd offset co_error_handler
               db dpl0
               dd offset alignment_check_handler
               db dpl0
               dd offset machine_check_exception
               db dpl0

end_of_initial_idt  equ $




;--------------------------------------------------------------------------
;
;       define idt gate
;
;--------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX    offset
;       BL     interrupt number
;       BH     dpl
;
;       DS     linear_space
;
;---------------------------------------------------------------------------

 
 assume ds:codseg
 
 
define_idt_gate:
 
  push  ebx
  push  edi

  movzx edi,bl
  shl   edi,3
  add   edi,offset idt

  shld  ebx,eax,16
  rol   ebx,16
  rol   eax,16
  mov   ax,phys_mem_exec
  rol   eax,16
  mov   bl,0
  add   bh,80h+intrgate

  mov   [edi],eax
  mov   [edi+4],ebx

  pop   edi
  pop   ebx

  ret


  icod  ends


;--------------------------------------------------------------------------
;
;       multi level interrupt switches
;
;--------------------------------------------------------------------------

  assume ds:nothing


  icode


initial_debug_exception_handler:


  ipre  debug_ec

  mov   al,debug_exception
  jmp   exception


  icod  ends




        align 16




multi_handler macro intr,icode

  align 4

intr&_handler:

  IFIDN <icode>,<ec_present>
        mov   byte ptr ss:[esp+3],hardware_ec
  ELSE
        push  icode
  ENDIF
  pushad
  mov   al,intr
  jmp   short multi_exception
  endm




  multi_handler divide_error,fault
  multi_handler breakpoint,trap1
  multi_handler overflow,fault
  multi_handler bound_check,fault
  multi_handler invalid_opcode,fault
  multi_handler double_fault,ec_present
  multi_handler stack_exception,ec_present
  multi_handler seg_not_present,ec_present
  multi_handler co_error,fault
  multi_handler alignment_check,ec_present




;----------------------------------------------------------------------------
;
;       std exception handler
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       stack like ipre, (multi exception: but seg regs not yet pushed!)
;
;       AL     intr number
;
;----------------------------------------------------------------------------






        align 16


  IF kernel_type NE pentium   ;----------------------------------------------


multi_exception:


exception:

  mov   ebp,esp
  and   ebp,-size tcb

  and   eax,000000FFh

  cmp   esp,PM
  jae   kd_exception


  CORZ  al,seg_not_present      ; ensures that seg regs are valid
  CORZ  al,stack_exception      ;
  IFZ   al,general_protection   ; recall: linear_space is only valid
                                ;         data segment
        mov   ecx,linear_space  ;
        mov   dx,ds             ;
        CORNZ edx,ecx           ;
        mov   dx,es             ;
        IFNZ  edx,ecx           ;
              mov   ds,cx       ;
              mov   es,cx       ;
              ipost             ;
                                ;
              align 16          ;
        FI                      ;
  FI                            ;


  sub    edi,edi
  IFAE   al,8
  CANDBE al,17
  CANDNZ al,16
         add  edi,4
  FI

  test  byte ptr [esp+ip_cs],11b
  jz    kd_exception


  ENDIF




  IF kernel_type EQ pentium   ;----------------------------------------------



multi_exception:

  push  ds


exception:

  mov   ebp,esp
  and   ebp,-size tcb

  and   eax,000000FFh

  cmp   esp,PM
  jae   kd_exception


  test  byte ptr [esp+ip_cs],11b
  jz    kd_exception


  CORZ  al,seg_not_present      ; ensures that seg regs are valid
  CORZ  al,stack_exception      ;
  IFZ   al,general_protection   ; recall: linear_space is only valid
                                ;         data segment
        mov   ecx,linear_space  ;
        CORNZ [esp+ip_ds],ecx   ;
        mov   dx,es             ;
        IFNZ  edx,ecx           ;
              mov   [esp+ip_ds],ecx
              mov   es,cx       ;
              ipost             ;
                                ;
              align 16          ;
        FI                      ;
  FI                            ;


  sub    edi,edi
  IFAE   al,8
  CANDBE al,17
  CANDNZ al,16
         add  edi,4
  FI


  ENDIF



  mov   ebp,[ebp+thread_idt_base]
  test  ebp,ebp
  jz    short perhaps_kd_exception


                                      ; note: define_pl3_idt ensures that
  lea   ebp,[eax*8+ebp]               ; idt_base is always valid
                                      ; (inside virtual_space)
  mov   ebx,[ebp+4]
  mov   bx,[ebp]

  test  ebx,ebx
  jz    short perhaps_kd_exception
  cmp   ebx,size virtual_space
  ja    short perhaps_kd_exception



  mov   ebp,[esp+ip_esp]

  bt    [esp+ip_eflags],vm_flag
  IFC
        ke    'v86_exc'
  FI
  ;;;;; jc    intr_from_v86

  sub   ebp,3*4
  jc    short perhaps_kd_exception
  sub   ebp,edi
  jc    short perhaps_kd_exception

 
  test  edi,edi
  IFNZ
        movzx edx,word ptr [esp+ip_error_code]
        mov   [ebp],edx
        add   ebp,edi
  FI

  mov   edx,[esp+ip_eip]
  mov   [ebp+iret_eip],edx
  mov   cx,[esp+ip_cs]
  mov   [ebp+iret_cs],cx
  mov   edx,[esp+ip_eflags]
  mov   [ebp+iret_eflags],edx

  btr   edx,t_flag
  mov   [esp+ip_eflags],edx
  mov   [esp+ip_eip],ebx
  mov   [esp+ip_cs],cx
  sub   ebp,edi
  mov   [esp+ip_esp],ebp

  ipost







perhaps_kd_exception:

  cmp   esp,max_kdebug_tcb
  ja    shutdown_thread



kd_exception:

  jmp   cs:[kdebug_exception+physical_kernel_info_page]







;----------------------------------------------------------------------------
;
;       emulate LIDT [EAX]
;
;----------------------------------------------------------------------------
; PRECONDITION:
;
;       EAX   instruction SHR 8
;       EDI   REG addr
;
;       DS    linear space
;
;----------------------------------------------------------------------------
; POSTCONDITION:
;
;       AL    instruction length
;
;----------------------------------------------------------------------------






emu_lidt_eax:

  mov   ebx,esp
  and   ebx,-size tcb

  mov   eax,[edi+7*4]

  CORA  eax,<size virtual_space-6>
  mov   eax,[eax+2]
  IFA   eax,<size virtual_space-32*8>
        sub   eax,eax
  FI

  mov   [ebx+thread_idt_base],eax

  mov   al,3
  ret





 
 
 
 

  code ends
  end
