include l4pre 

  Copyright GMD, L4.SGMCTR, 03 07 96, 19
 
 
;*********************************************************************
;******                                                         ******
;******         Segment Controller                              ******
;******                                                         ******
;******                                   Author:   J.Liedtke   ******
;******                                                         ******
;******                                   created:  24.02.90    ******
;******                                   modified: 03.07.96    ******
;******                                                         ******
;*********************************************************************
 
  public init_sgmctr
 


  extrn configuration:byte


 
include l4const
include adrspace
include uid
include tcb
include cpucb


ok_for 486,pentium,ppro



;-------------------------------------------------------------------------
;
;        descriptor types
;
;-------------------------------------------------------------------------
 
 
r32               equ  0100000000010000b
rw32              equ  0100000000010010b
r16               equ             10000b
rw16              equ             10010b
 
x32               equ  0100000000011000b
xr32              equ  0100000000011010b
x16               equ             11000b
xr16              equ             11010b
 
ldtseg            equ  2
taskgate          equ  5
tsseg             equ  9
callgate          equ  0Ch
intrgate          equ  0Eh
trapgate          equ  0Fh
 
 
;---------------------------------------------------------------------------
;
;        descriptor privilege levels codes
;
;---------------------------------------------------------------------------
 
dpl0    equ 0 shl 5
dpl1    equ 1 shl 5
dpl2    equ 2 shl 5
dpl3    equ 3 shl 5



;----------------------------------------------------------------------------
;
;       descriptor entry
;
;----------------------------------------------------------------------------


descriptor macro dtype,dpl,dbase,dsize


IF dsize eq 0

  dw    0FFFFh
  dw    loww        dbase
  db    low highw   dbase
  db    low         (dtype+dpl+80h)
  db    high        (dtype+8000h) + 0Fh
  db    high highw  dbase

ELSE
IF dsize AND -KB4

  dw    loww        ((dsize SHR 12)-1)
  dw    loww        dbase
  db    low highw   dbase
  db    low         (dtype+dpl+80h)
  db    high        (dtype+8000h) + highw ((dsize SHR 12)-1)
  db    high highw  dbase

ELSE

  dw    loww        (dsize-1)
  dw    loww        dbase
  db    low highw   dbase
  db    low         (dtype+dpl+80h)
  db    high        dtype
  db    high highw  dbase

ENDIF
ENDIF

  endm


 
 

;****************************************************************************
;******                                                               *******
;******                                                               *******
;******                 Segment Controller Initialization             *******
;******                                                               *******
;******                                                               *******
;****************************************************************************



;-----------------------------------------------------------------------
;
;       init segment controller
;
;-----------------------------------------------------------------------
; PRECONDITION:
;
;       protected mode
;
;       paging enabled, adrspace established
;
;       disable interrupt
;
;       DS : R/W 0..4GB
;       SS : R/W 0..4GB
;       CS : X/R 0..4GB, USE32
;
;-----------------------------------------------------------------------
; POSTCONDITION:
;
;       GDT     initialized
;       GDTR    initialized
;
;       LDTR    initialized
;
;       CS    phys_mem_exec
;       DS    linear_space
;       ES    linear_space
;       FS    linear_space
;       GS    linear_space
;       SS    linear_space
;
;       EAX...EBP scratch
;
;----------------------------------------------------------------------
 

 assume ds:codseg

  icode

 
 
init_sgmctr:

  mov   ax,ds
  mov   es,ax

  mov   edi,offset gdt + first_kernel_sgm
  mov   ecx,(size gdt - first_kernel_sgm)/4
  sub   eax,eax
  cld
  rep   stosd

  mov   edi,offset gdt + first_kernel_sgm
  mov   esi,offset initial_gdt+8
  mov   ecx,(offset end_of_initial_gdt - (offset initial_gdt+8) +3) / 4
  rep   movsd

  lgdt  pword ptr ds:[gdt_vec]

  jmpf32 $+6,phys_mem_exec

  mov   ax,linear_kernel_space
  mov   ds,ax
  mov   es,ax
  mov   fs,ax
  mov   gs,ax
  mov   ss,ax
  lea   esp,[esp+PM]

  mov   ax,std_ldt_seg
  lldt  ax

  ret
 
 
        align 4


gdt_vec       dw size gdt-1
              dd offset gdt

        align 4

  IF kernel_type NE pentium

        user_space_size        equ     size linear_address_space
  ELSE
        user_space_size        equ     size virtual_space + MB4

  ENDIF



.errnz  size virtual_space AND (MB4-1)


.xall
initial_gdt   dd 0,0                   ; dummy seg

  descriptor rw32, dpl0, 0, <size linear_address_space>   ; 08 : linear_kernel_space
  descriptor rw32, dpl3, 0,  user_space_size              ; 10 : linear space
  descriptor xr32, dpl3, 0,  user_space_size              ; 18 : linear space

  descriptor rw32, dpl2, PM, <size physical_kernel_memory>; 20 : phys_mem
  descriptor xr32, dpl0, PM, <size physical_kernel_memory>; 28 : phys_exec

  descriptor ldtseg, dpl0, <offset ldt>, <size ldt>       ; 30 : LDT

  tss_base    equ offset cpu_tss_area
  tss_size    equ offset iopbm - offset cpu_tss_area + size iopbm

  descriptor tsseg, dpl0, tss_base, tss_size              ; 38 : cpu0_tss


end_of_initial_gdt  equ $


  icod  ends


  code ends
  end





